# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dccmd',
 'dccmd.main',
 'dccmd.main.auth',
 'dccmd.main.crypto',
 'dccmd.main.download',
 'dccmd.main.models',
 'dccmd.main.upload',
 'dccmd.main.users',
 'dccmd.main.util']

package_data = \
{'': ['*']}

install_requires = \
['SecretStorage>=3.3.1,<4.0.0',
 'dracoon>=1.5.3,<2.0.0',
 'keyring>=23.6.0,<24.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['dccmd = dccmd:app']}

setup_kwargs = {
    'name': 'dccmd',
    'version': '0.3.4',
    'description': 'DRACOON Commander – CLI client for DRACOON Cloud (dracoon.com)',
    'long_description': '<h1 align="center">DRACOON Commander</h1>\n\n\n## Table of Contents\n\n* [About the Project](#about-the-project)\n  * [Built With](#built-with)\n* [Getting Started](#getting-started)\n  * [Prerequisites](#prerequisites)\n  * [Installation](#installation)\n* [Usage](#usage)\n* [Configuration](#configuration)\n* [License](#license)\n\n\n## About the project\n_Disclaimer: This is an unofficial client and is not supported by DRACOON._<br>\nThis client is a CLI tool to perform basic commands in DRACOON and comes with the following functionalities:\n\n* Full support for S3 direct up- / download\n    * Chunked up- and downloads\n* Full support for DRACOON end-to-end encryption\n* Optimized for concurrent requests\n* Store credentials in OS-specific secure location\n    * Linux: Freedesktop Secret Service (secretstorage)\n    * macOS: Keychain\n    * Windows: Windows Credential Locker\n\n### Built With\n* [typer](https://typer.tiangolo.com)\n* [keyring](https://pypi.org/project/keyring)\n* [dracoon](https://github.com/unbekanntes-pferd/dracoon-python-api)\n    * [httpx](https://www.python-httpx.org/)\n* [poetry](https://python-poetry.org/)\n\nA full dependency list can be viewed in\n* [pyproject.toml](/pyproject.toml) - list of dependencies and project info\n* [poetry.lock](/poetry.lock)\n\nDRACOON Commander is built with typer as the CLI framework and uses keyring to store all credentials (OAuth2 tokens, client credentials and encryption password). \nThe tool is built on top of dracoon, an async API wrapper for DRACOON based on httpx.\nThe project is managed with poetry (dependencies, release build and publishing).\n\n## Getting Started\nIn order to get started, download the latest tarball from Github or install dccmd from pip:\n[Releases]()\n\n### Prerequisites\nYou need a working Python 3.10 installation – dccmd makes use of type annotations and uses 3.10 features.\nGet the latest Python version from: [python.org](https://python.org).\n\n```bash\npython3 --version\n```\n\nIn order to get going, you can install dccmd either in a virtual environment or globally.\n\n### Installation\n\nTo install a version, use `pip` and install the `dccmd` package.\n\n#### In a virtual environment\n```bash\nvirtualenv <DIR>\nsource <DIR>/bin/activate \npython3 -m pip install dccmd\n```\n#### Globally\n```bash\npython3 -m pip install dccmd\n```\n\n#### Set PATH\nIn order for the script to work, you might need to add the relevant script path to your PATH.\nWhen installing with pip, the output will already indicate if the path is present or not.\nIf you do not add the correct directory to PATH, you will *not* be able to use the `dccmd` command in your preferred shell.\n\n##### Windows\nIn Windows, just add the script path by editing the environment variables for your account:\nLook for an entry called \'Path\' and install the script directory from the `pip install` output.\n\n##### Unix\nOn Linux or macOS you can add a path to PATH by using the following command:<br>\n`export PATH="/your/directory/see/install/output:$PATH"`<br>\n\n\n## Usage\n\n### Display commands\n\nIn order to see all available commands, arguments and options, use the --help flag:\n```bash\ndccmd --help\n```\n\n```\nUsage: dccmd [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n  --help                          Show this message and exit.\n\nCommands:\n  auth\n  client\n  crypto\n  download  Download a file from DRACOON by providing a source path and a...\n  ls        List all nodes in a DRACOON path\n  mkdir     Create a folder in a DRACOON parent path\n  mkroom    Create a room (inherit permissions) in a DRACOON parent path\n  rm        Delete a file / folder / room in DRACOON\n  upload    Upload a file into DRACOON by providing a source path and a...\n```\nAll commands display their own help message, e.g. \n`dccmd upload --help`.\n\n### Client registration and authentication\n\nBefore you can perform any command, you must authenticate and set up the client.\nIf you enter any command which requires authentication (e.g. `dccmd ls your.dracoon.domain.com/`), you will be prompted first for a client configuration:\n* client id\n* client secret\n\n#### Client \nBefore you can use `dccmd` you need to generate a client in your DRACOON instance (config manager role required).\n1. Create a client with a client id and client secret.\nPlease make sure you have the following settings active:\n* Authorization code \n* Redirect URI is set to `https://your.dracoon.domain.com/oauth/callback`\n* Optional: If you wish to use the CLI mode (enter password and username via CLI), you can activate password flow \n\n2. Copy client id and client secret and use any command (e.g. `dccmd ls your.dracoon.domain.com/`).\n\n3. Enter client id and client secret – the information will be securely stored in your OS-specific secret container.\n\n#### Authentication\nOnce the client is set up, you will receive a link to authenticate via OAuth2 authorization code flow – you will then receive a code which you need to enter into the terminal.\nWhen completed, you will be prompted to store credentials securely (OS-specific).\n\nAdditionally, you can skip the authorization code flow and provide credentials directly, e.g. for the `dccmd ls` command:\n\n```bash\ndccmd ls your-dracoon.domain.com/ --cli-mode username@mail.com topsecret123!\n```\n\n### Upload\n\n**Important: if you use Windows, you need to provide the path with \'/\' instead of \'\\\\\'!**\n\nYou can upload single files using the upload command:\n\n```bash\ndccmd upload /path/to/file.pdf your-dracoon.domain.com/\n```\nIn order to upload a directory, use the `--recursive` (`-r`) flag:\n\n```bash\ndccmd upload -r /path/to/folder your-dracoon.domain.com/\n```\n\n#### Conflict resolution\nIf you upload a file which already exists (based on file name), the upload will be rejected.<br> \nIn order to force an overwrite, use the `--overwrite` flag:\n\n```bash\ndccmd upload /path/to/file.pdf your-dracoon.domain.com/ --overwrite\n```\n\nIf you wish to auto-rename the file if it already exists, use the `auto-rename`flag:\n\n```bash\ndccmd upload /path/to/file.pdf your-dracoon.domain.com/ --auto-rename\n```\n\n#### Advanced usage\nIf you upload folders recursively, you might encounter performance issues, specifically when uploading many small files. \nYou can therefore adjust concurrent file uploads via the `--velocity` (`-v`) flag:\n\n```bash\ndccmd upload -r /path/to/folder your-dracoon.domain.com/ -v 3\n```\nThe default value is 2 - it does not coincide with real request value.<br>\nMaximum (although not recommended) value is 10. Entering higher numbers will result in max value use.<br>\nMinimum value is 1 - this will not upload a folder per file but is the minimum concurrent request value. Entering lower numbers will result in min value use.\n\nIf you need to understand why uploads fail, you can also run the command using the `--debug` flag:\n\n```bash\ndccmd upload -r /path/to/folder your-dracoon.domain.com/ --debug\n```\n*Note: This will have impact on performance as the log will be streamed to terminal and the log level will be increased to DEBUG.*\n\n### Create folder\n\nIf you wish to create a folder, use the `mkdir` command:\n\n```bash\ndccmd mkdir your-dracoon.domain.com/parent/newfolder\n```\nJust enter the full new path to create a folder. \nYou will need *create* permission to do so.\n\n\n### Create room\n\nIf you wish to create a room, use the `mkroom` command:\n\n```bash\ndccmd mkroom your-dracoon.domain.com/parent/newroom\n```\nJust enter the full new path to create a room. \nThe room will be created as a room with inherited permissions from the parent.\nYou will need *manage* permission to do so.\n\nIt is currently not possible to create a room in the root path.\n\n### Delete node\n\nIf you wish to delete a node, use the `rm` command:\n\n```bash\ndccmd rm your-dracoon.domain.com/parent/somefile.pdf\n```\nIn order to delete a container (room, folder) you need to use the `--recursive` (`r`) flag:\n\n```bash\ndccmd rm your-dracoon.domain.com/parent/folder/to/delete\n```\n**Warning: Deleting rooms cannot be undone!**\n\n### List nodes\n\nIn order to list all nodes, use the `ls` command:\n\n```bash\ndccmd ls your-dracoon.domain.com/\n```\n*Note: In order to list the root node, you need to provide a trailing `/`*\nFor a specific container (room or folder), use the path:\n```bash\ndccmd ls your-dracoon.domain.com/your/room\n```\n\n#### Displaying additional information\nUsing the `ls` command by default only provides node names.\nIn order to display more information, use relevant flags:\n\n* Display all information (size, last updated, last update user): `--long` (`-l`)\n    * Display sizes in human readable format (B, KB..): `--human-readable` (`-h`)\n* Display node id: `--inode` (`-i`)\n\nExample displaying full information:\n\n```bash\ndccmd ls -h -i -l your-dracoon.domain.com/your/room\n```\n### Download\n\nTo download a file, use the `download` command:\n\n```bash\ndccmd download your-dracoon.domain.com/your/cool-file.mp4 /target/directory\n```\n\nTo download a room or a folder, use the `download` command with `--recursive` (`-r`) flag:\n\n```bash\ndccmd download -r your-dracoon.domain.com/your/cool-folder /target/directory\n```\n\n#### Advanced usage\nIf you download folders recursively, you might encounter performance issues, specifically when downloading many small files. \nYou can therefore adjust concurrent file uploads via the `--velocity` (`-v`) flag:\n\n```bash\ndccmd upload -r /path/to/folder your-dracoon.domain.com/ -v 3\n```\nThe default value is 2 - it does not coincide with real request value.<br>\nMaximum (although not recommended) value is 10. Entering higher numbers will result in max value use.<br>\nMinimum value is 1 - this will not download a folder per file but is the minimum concurrent request value. Entering lower numbers will result in min value use.\n\n### User operations\n\nYou can list, edit and import users with relevant `dccmd users` command:\n\n* csv-import  Add a list of users to DRACOON from a CSV file\n* ls          Get a list of users in DRACOON\n* rm          Delete a user\n\n#### Importing users\n\nYou can import users by using the `csv-import` command and providing a path to the csv file:\n\n```bash\ndccmd users csv-import /path/to/users.csv your-dracoon.domain.com/\n```\n\nThe csv file must contain a header and should include the following attributes:\n\n* first name\n* last name\n* email \n* login (optional)\n\nBy default, local users are created - if you want to import oidc users, you need pass the oidc config id:\n\n```bash\n#example with OIDC config 5\ndccmd users csv-import /path/to/users.csv your-dracoon.domain.com/ 5\n```\n\n#### Listing users\n\nYou can list all users using the `ls` command:\n\n```bash\ndccmd users ls your-dracoon.domain.com/\n```\n\nYou can get all users also as csv format by using the `--csv` flag:\n\n```bash\ndccmd users ls your-dracoon.domain.com/ --csv > users.csv\n```\n\nTo find a user, you can pass a search string to search for either first name, last name or user name (search string applies to all):\n\n```bash\n# will return all users with either first name, last name or user name containing \'yourname\'\ndccmd users ls your-dracoon.domain.com/ yourname\n```\n\n#### Deleting users\n\nYou can delete a user by providing the username:\n```bash\ndccmd users rm your-dracoon.domain.com/ user123\n```\n\n\n## Configuration / administration\n\nYou can view / manage the configuration for `dccmd` using the relevant commands:\n\n* `dccmd auth` - manage credentials\n    * `dccmd auth ls your.dracoon.domain.com` will display if a refresh token has been stored for the provided domain\n    * `dccmd auth rm your-dracoon.domain.com` will remove stored credentials for the provided domain\n* `dccmd client` - manage client\n    * `dccmd client register your.dracoon.domain.com` will start the registration process for a client and given domain\n    * `dccmd client ls your.dracoon.domain.com` will display client information for the provided domain\n    * `dccmd client rm your-dracoon.domain.com` will remove the stored client config for the provided domain\n* `dccmd crypto` - manage encryption\n    * `dccmd crypto ls your.dracoon.domain.com` will display if encryption password is stored for the provided domain\n    * `dccmd crypto rm your-dracoon.domain.com` will remove the encryption password for the provided domain\n    * `dccmd crypto distribute your-dracoon.domain.com/` will generate file keys available to distribute - if providing a specific path (`dccmd crypto distribute your-dracoon.domain.com/some/path`), only keys for provided parent room will be generated.\n\n### Logging \nWhen using a command, a log will be created in the current working directory.\nCurrently it is not possible to configure a default path for a log.\n\nYou can stream the log to stdout by using the `--debug` flag with any DRACOON-specific command (`upload`, `download`, `ls`, `rm`, `mkdir`, `mkroom`).\n\n\n## License\nDistributed under the Apache License. See [LICENSE](/LICENSE) for more information.\n',
    'author': 'Octavio Simone',
    'author_email': '70800577+unbekanntes-pferd@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
